ETRX2USB Sender
Copyright (C) 2007 Joseph Bosman, Steve Olivieri, Ipek Ozil, Brandon Steacy

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA


#include "em250.h"

HANDLE setupCOM(LPCSTR port)
{
	/* Serial port configuration settings. */
	DCB comConfig;
	
	/* Serial port timeout settings. */
	COMMTIMEOUTS comTimeouts;
	
	/* The file handle to return to the caller. */
	HANDLE ret;
	
	/* Open the serial port. */
	ret = CreateFile((LPCSTR)port, GENERIC_READ|GENERIC_WRITE, 0, NULL,
		OPEN_EXISTING, 0, NULL);
	if(ret == INVALID_HANDLE_VALUE)
		return INVALID_HANDLE_VALUE;

	/* Get the port's configuration settings, then change them to better
	 * match the ETRX2USB device.
	 */
	if(GetCommState(ret, &comConfig))
	{
		comConfig.BaudRate = 19200;
		comConfig.ByteSize = 8;
		comConfig.Parity = NOPARITY;
		comConfig.StopBits = ONESTOPBIT;
		comConfig.fBinary = FALSE;
		comConfig.fParity = FALSE;
	}
	else
		return INVALID_HANDLE_VALUE;
	if(!SetCommState(ret, &comConfig))
		return INVALID_HANDLE_VALUE;

	/* Get the port's timeout settings, then set them all to 25 ms.  This
	 * is probably not an ideal solution.
	 */
	if(GetCommTimeouts(ret, &comTimeouts))
	{
		comTimeouts.ReadIntervalTimeout = 25;
		comTimeouts.ReadTotalTimeoutConstant = 25;
		comTimeouts.ReadTotalTimeoutMultiplier = 25;
		comTimeouts.WriteTotalTimeoutConstant = 25;
		comTimeouts.WriteTotalTimeoutMultiplier = 25;
	}
	else
		return INVALID_HANDLE_VALUE;
	if(!SetCommTimeouts(ret, &comTimeouts))
		return INVALID_HANDLE_VALUE;

	/* Empty the input and output buffers for the serial port. */
	PurgeComm(ret, PURGE_RXCLEAR|PURGE_TXCLEAR);
	
	/* Return the file handle for the serial port. */
	return ret;
}

int writeSerial(LPCSTR data, PHANDLE comPort, UINT len, BOOL command)
{
	/* Single-character buffer for writing strings to the serial port. */
	CHAR zBuff;
	
	/* How many bytes we wrote on the last write call. */
	DWORD bWritten;
	
	/* How many bytes we've read. */
	DWORD bRead;
	
	/* Loop variable. */
	UINT i;
	
	/* Write a command to the serial port. */
	for(i = 0; i < len; i++)
	{
		zBuff = *(data+i);
		if(!WriteFile(*comPort, (LPVOID)&zBuff, 1, &bWritten, NULL))
			return FAILURE;
	}
	
	/* Send a carriage return to finish the command. */
	if(command)
	{
		zBuff = '\r';
		if(!WriteFile(*comPort, (LPVOID)&zBuff, 1, &bRead, NULL))
			return FAILURE;
	}

	return SUCCESS;
}

int readMessage(PHANDLE in, LPSTR buff)
{
	/* Temporary buffer. */
	CHAR zBuff;
	
	/* Stores how many bytes the last ReadFile() read. */
	DWORD bRead;
	
	/* Stores how many total bytes we've read. */
	INT i;
	
	/* Initialize i.  Then, read in the data one byte at a
	 * time and write it to buff.  Data is read this way for COM
	 * support.  We finish when there is no more data to be read.
	 */
	i = 0;
	do
	{
		ReadFile(*in, (LPVOID)&zBuff, 1, &bRead, NULL);
		if(bRead > 0)
		{
			*(buff+i) = zBuff;
			++i;
		}
	} while (bRead > 0);

	return i;
}

LPSTR getMessage(PHANDLE in, LPCSTR message, PHANDLE out)
{
	/* Temporary buffer for storing data that we read from the COM port. */
	LPSTR buff;
	
	/* Temporary buffer that points to part of buff containing the desired
	 * message (if it exists) or nothing (otherwise).
	 */
	LPSTR tBuff;
	
	/* We copy tBuff into a separate buffer for returning to avoid memory
	 * leaks.
	 */
	LPSTR retBuff;
	
	/* How many bytes we wrote. */
	DWORD bWritten;
	
	/* How many bytes we read. */
	INT i;

	/* Since we're storing temporary data in buff, we need to allocate
	 * memory for it now.  Since we use tBuff is a BOOL, we leave it
	 * NULL for now and only give it a value when we find the message
	 * that we're looking for (or an ERROR).
	 */
	buff = (LPSTR)calloc(150, sizeof(CHAR));
	tBuff = NULL;

	/* While we haven't found the desired message (or an ERROR), keep
	 * looking!
	 */
	while(!tBuff)
	{
		/* Read data from the COM port. */
		memset(buff, 0, 150);
		i = readMessage(in, buff);
		
		/* Search that data for our desired message. If we don't find it
		 * search for an ERROR message.
		 */
		tBuff = strstr(buff, message);
		if(!tBuff)
			tBuff = strstr(buff, "ERROR:");
			
		/* If out is not NULL, write any data that we've read to the file
		 * pointed to by out.  Usually, this is a log file.  And, we don't
		 * really care if it fails.
		 */
		if(out)
			WriteFile(*out, (LPVOID)buff, i, &bWritten, NULL);
	}
	
	/* Since we have something to return, let's allocate memory for retBuff
	 * and then copy the useful data into it.  The caller of getMessage()
	 * is responsible for freeing retBuff.
	 */
	retBuff = (LPSTR)calloc(150, sizeof(CHAR));
	strcpy(retBuff, tBuff);

	free(buff);
	free(tBuff);
	return retBuff;
}

DWORD ceiling(double x)
{
	/* Return value. */
	int ret;
	
	/* This basically abuses casting in C.   If I cast a float (or a double)
	 * to an integer, anything after the decimal point is simply dropped.
	 */
	ret = (DWORD)x;
	
	/* Now, if the chopped version is equal to the original, then we were
	 * looking at an integer to begin with.  Otherwise, there was a decimal
	 * component and we need to round up.
	 */
	if(ret == x)
		return ret;
	else
		return (ret + 1);
}

int main(int argc, char* argv[])
{
	/* Handle for the COM port we're writing to. */
	HANDLE serial;
	
	/* Handle for the file we're sending. */
	HANDLE imageFile;
	
	/* The COM port to connect to. */
	LPSTR comPort;
	
	/* Filename for the input file. */
	LPSTR fileName;
	
	/* Buffers for storing data. */
	LPSTR buff;
	LPSTR tBuff;
	
	/* The hardware ID number for the sender. */
	LPSTR hwID;
	
	/* How many bytes we read. */
	DWORD bRead;
	
	/* The size of the file.  Note that we don't support files
	 * larger then 2GB.
	 */
	DWORD fileSize;
	
	/* The number of packets that we need to send. */
	DWORD numPackets;
	
	/* Loop variables. */
	INT i;
	UINT j;
	
	/* Make sure that they specify the required command
	 * arguments (COM port and directory).
	 */
	if(argc < 5)
		return MAIN_INPUT_ERROR;
	
	/* Allocate memory for our strings. */
	comPort = (LPSTR)calloc(6, sizeof(CHAR));
	fileName = (LPSTR)calloc(1024, sizeof(CHAR));
	buff = (LPSTR)calloc(150, sizeof(CHAR));
	hwID = (LPSTR)calloc(17, sizeof(CHAR));
	
	/* Initialize variables. */
	tBuff = NULL;
	bRead = 0;
	numPackets = 0;
	fileSize = 0;
	
	/* Parse input. */
	for(i = 1; i < argc; i++)
	{
		if(strcmp(argv[i], "-c") == 0)
			sprintf(comPort, "COM%s:", argv[++i]);
		else if(strcmp(argv[i], "-f") == 0)
			sprintf(fileName, "%s", argv[++i]);
		else
			return MAIN_INPUT_ERROR;
	}

	/* Open the serial port. */
	serial = setupCOM((LPCSTR)comPort);
	if(serial == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "Could not open %s.\n", comPort);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_SERIAL_PORT_FAIL;
	}

	/* Tell the EM250 to leave any PANs that it might be in.   If we get an
	 * ERROR here, that just means that we weren't already in a PAN.  So, we
	 * can ignore it.
	 */
	if(writeSerial("AT+DASSL", &serial, strlen("AT+DASSL"), TRUE))
	{
		fprintf(stderr, "Could not write to %s.\n", comPort);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_DATA_WRITE_ERROR;
	}
	tBuff = getMessage(&serial, "LeftPAN", NULL);
	free(tBuff);

	/* Turn off the annoying echo when we're sending binary data.
	 * Failure to do so will break ACKs and other fun things.
	 */
	if(writeSerial("ATS0B4=1", &serial, strlen("ATS0B4=1"), TRUE))
	{
		fprintf(stderr, "Could not write to %s.\n", comPort);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_DATA_WRITE_ERROR;
	}
	tBuff = getMessage(&serial, "OK", NULL);
	if(strstr(tBuff, "ERROR:"))
	{
		fprintf(stderr, "Could not set register value 0B4, %s.\n", tBuff);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		free(tBuff);
		return MAIN_REGISTER_FAIL;
	}
	free(tBuff);
	
	/* We also need to turn off the ability to have multiple packets in
	 * flight at once.  That way, we're certain that the receiver will
	 * never be overloaded.
	 */
	if(writeSerial("ATS06D=1:TG-ETRX1", &serial, strlen("ATS06D=1:TG-ETRX1"), TRUE))
	{
		fprintf(stderr, "Could not write to %s.\n", comPort);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_DATA_WRITE_ERROR;
	}
	tBuff = getMessage(&serial, "OK", NULL);
	if(strstr(tBuff, "ERROR:"))
	{
		fprintf(stderr, "Could not set register value 06D, %s.\n", tBuff);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		free(tBuff);
		return MAIN_REGISTER_FAIL;
	}
	free(tBuff);
	
	/* Join the next best ZigBee network. */
	if(writeSerial("AT+JN", &serial, strlen("AT+JN"), TRUE))
	{
		fprintf(stderr, "Could not write to %s.\n", comPort);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_DATA_WRITE_ERROR;
	}
	tBuff = getMessage(&serial, "JPAN:", NULL);
	if(!strstr(tBuff, "JPAN:"))
	{
		fprintf(stderr, "Could not join a PAN.\n");
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		free(tBuff);
		return MAIN_JOIN_FAIL;
	}
	strncpy(buff, tBuff+5, 7);
	fprintf(stdout, "Joined network %s.\n", buff);
	free(tBuff);
	memset(buff, 0, 150);
	
	/* There's almost certainly a bunch of data in the buffers
	 * now, but we don't really care what it says.  So, we'll
	 * purge it.
	 */
	PurgeComm(&serial, PURGE_RXCLEAR);
	
	/* Now, we need to get the hardware ID of our coordinator.  The
	 * coordinator always acts as the base station, so this is where
	 * we want to send our files.
	 */
	if(writeSerial("AT+SN:00", &serial, strlen("AT+SN:00"), TRUE))
	{
		fprintf(stderr, "Could not write to %s.\n", comPort);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_DATA_WRITE_ERROR;
	}
	tBuff = getMessage(&serial, "COO:", NULL);
	if(!strstr(tBuff, "COO:"))
	{
		fprintf(stderr, "Could not find PAN coordinator.\n");
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		free(tBuff);
		return MAIN_COO_ERROR;
	}
	strncpy(hwID, tBuff+4, 16);
	free(tBuff);
	fprintf(stdout, "PAN Coordinator: %s\n", hwID);
	
	/* Open the image that we plan to send.  We do this before sending
	 * the IMAGE_READY signal because we first need to calculate how
	 * many packets we'll be sending.
	 */
	imageFile = CreateFile((LPCSTR)fileName, GENERIC_READ, FILE_SHARE_READ,
		NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if(imageFile == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "Could not open file: %s\n", fileName);
		CloseHandle((PVOID)&serial);
		free(comPort);
		free(fileName);
		free(buff);
		free(hwID);
		return MAIN_READ_ERROR;
	}
	fprintf(stdout, "Opened file: %s\n", fileName);
	
	/* Get the size of the file.  Note that we're limited to files of 2GB
	 * or less because we're using GetFileSize() instead of GetFileSizeEx().
	 * This should be okay, as I don't expect a 640x480 JPEG to ever be more
	 * than 2GB.  In fact, it can't be!
	 */
	fileSize = GetFileSize(imageFile, NULL);
	
	/* Our packets will each contain 50 bytes of data.  So, determine
	 * how many packets we'll need to send by diving the file size by 50
	 * and taking the ceiling of that number.
	 */
	numPackets = ceiling((double)((fileSize)/65.0f));
	fprintf(stdout, "File is %li bytes, need to send %li packets.\n",
		fileSize, numPackets);
	
	/* Now, we'll let the base station know that we have a file to send.  If
	 * we get the IMAGE_READY signal through, we'll wait for IMAGE_READY_OK.
	 * If we don't get it, we'll send IMAGE_READY again and wait again.  We'll
	 * do this at most 10 times before giving up entirely.
	 */
	for(j = 0; j < 10; j++)
	{
		/* Send the IMAGE_READY signal to the base station.  The format for
		 * this command is IMAGE_READY0xNNNN where NNNN is numPackets in
		 * hex.  We'll try this 10 times before giving up.
		 */
		memset(buff, 0, 150);
		sprintf(buff, "AT+UCAST:%s=IMAGE_READY0x%.4x",
			hwID, (unsigned int)numPackets);
		for(i = 0; i < 10; i++)
		{
			writeSerial(buff, &serial, strlen(buff), TRUE);
			fprintf(stdout, "Sent IMAGE_READY0x%.4x to %s.\n",
				(unsigned int)numPackets, hwID);
				
			/* Since we set register S06D, we'll get either "OK" or "ERROR"
			 * in response to our transmission (rather than ACK/NACK).
			 */
			tBuff = getMessage(&serial, "OK", NULL);
			if(strstr(tBuff, "OK"))
			{
				fprintf(stdout, "\tGot an ACK from %s.\n", hwID);
				break;
			}
			else
			{
				fprintf(stdout, "\tNo ACK from %s.\n", hwID);
				
				/* If we've already tried ten times (i == 9), then set
				 * i = 10 for later error checking.
				 */
				if(i == 9)
					++i;
				free(tBuff);
			}
		}
		
		/* If none of our ten attempts worked, we'll have to quit. */
		if(i == 10)
		{
			fprintf(stderr, "Failed to send IMAGE_READY signal to %s.\n",
				hwID);
			CloseHandle((PVOID)&serial);
			CloseHandle((PVOID)&imageFile);
			free(comPort);
			free(fileName);
			free(buff);
			free(hwID);
			return MAIN_NO_CONNECTION;
		}
		
		/* Now, we need to wait for the IMAGE_READY_OK signal. Sometimes it
		 * gets read in with the ACK.  So, check for it first.  If we don't
		 * already have it, wait for it.  If we *still* don't get it, try
		 * again.
		 */
		if(!strstr(tBuff, "IMAGE_READY_OK"))
		{
			free(tBuff);
			tBuff = getMessage(&serial, "IMAGE_READY_OK", NULL);
		}
		if(!strstr(tBuff, "IMAGE_READY_OK"))
		{
			fprintf(stdout, "Did not get the OK from %s\n", hwID);
			continue;
		}
		else
			break;
	}
	
	/* If we made it here, we've got the IMAGE_READY_OK signal. */
	fprintf(stdout, "Got the IMAGE_READY_OK signal from %s.\n", hwID);
	
	/* No, we can begin sending packets. */
	for(i = 0; i < numPackets; i++)
	{
		/* Read the next 50 bytes in the file. */
		memset(buff, 0, 150);
		ReadFile(imageFile, (LPVOID)buff, 65, &bRead, NULL);
		
		/* Now, try to send it to the base station.  We'll try each
		 * packet at most 10 times before aborting.
		 */
		for(j = 0; j < 10; j++)
		{
			/* Send the command. */
			tBuff = (LPSTR)calloc(150, sizeof(CHAR));
			sprintf(tBuff, "AT+UCASTB:%.2x,%s", (unsigned int)bRead, hwID);
			writeSerial(tBuff, &serial, strlen(tBuff), TRUE);
			
			/* Just to be safe. */
			Sleep(10);
			
			/* Send the data. */
			writeSerial(buff, &serial, bRead, FALSE);
			free(tBuff);
			fprintf(stderr, "Sent packet %i of %li (%li bytes).\n", i+1, numPackets, bRead);
			
			/* Wait for OK/ERROR. */
			tBuff = getMessage(&serial, "OK", NULL);
			if(strstr(tBuff, "OK"))
			{
				fprintf(stdout, "\tGot ACK.\n");
				free(tBuff);
				PurgeComm(&serial, PURGE_TXCLEAR|PURGE_RXCLEAR);
				break;
			}
			else
			{
				free(tBuff);
				PurgeComm(&serial, PURGE_TXCLEAR|PURGE_RXCLEAR);
				
				/* If we've already tried 10 times, give up. */
				if(j == 9)
				{
					++j;
					break;
				}
			}
		}
		
		/* We failed 10 times, so we're quitting. */
		if(j == 10)
		{
			fprintf(stderr, "Lost connection to %s, aborting.\n", hwID);
			break;
		}
	}
	
	fprintf(stdout, "Sent %i of %li packets to %s.\n", i, numPackets, hwID);
	
	/* Close the files. */
	CloseHandle((PVOID)&serial);
	CloseHandle((PVOID)&imageFile);
	
	/* Free any used memory and exit. */
	free(comPort);
	free(fileName);
	free(buff);
	free(hwID);

	return MAIN_SUCCESS;
}
